/*
 * Copyright (C) 2005-present, 58.com.  All rights reserved.
 * Use of this source code is governed by a BSD type license that can be
 * found in the LICENSE file.
 */

import 'dart:convert';
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context.dart';
import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/analysis/session.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/visitor.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:flutter/material.dart' hide Element;
import 'package:flutter_test/flutter_test.dart';
import 'package:path/path.dart';

import 'private_transformer.dart';

// 完善基类判断可以减少递归查找，提高效率
var baseWidget = [
  'StatefulWidget',
  'StatelessWidget',
  'Widget',
  'CustomScrollView',
  'ScrollView',
  'ListView',
  'GridView',
];

var baseAPI = [
  'Alignment',
  'EdgeInsets',
  'TextStyle',
  'BoxDecoration',
  'DecorationImage',
  'NetworkImage',
  'Rect',
  'OffsetBase',
  'Radius',
  'RRect',
  'RSTransform',
  'BorderRadiusGeometry',
  'RenderComparison',
  'Axis',
  'VerticalDirection',
  'AxisDirection',
];
const shrinking = {
  'cupertino': 'c',
  'widgets': 'w',
  'painting': 'p',
  'material': 'm',
  'rendering': 'r',
  'a': 'a',
};
var flutterSDK = '/Users/aven/Android/flutter';
var frameworkAPI = [
  'packages/flutter/lib/src/widgets/',
  'packages/flutter/lib/src/painting/',
  'packages/flutter/lib/src/material/',
  'packages/flutter/lib/src/cupertino/',
  'packages/flutter/lib/src/rendering/'
];

void main() {
  TestWidgetsFlutterBinding.ensureInitialized();

  test('Generate widget binding from widget source code', () async {
    List<String> includedPaths = <String>[
      //'$flutterSDK/bin/cache/pkg/sky_engine/lib/ui/geometry.dart'
    ];
    for (var dirPath in frameworkAPI) {
      var dir = Directory('$flutterSDK/$dirPath');
      if (!await dir.exists()) continue;
      var files = await dir
          .list(recursive: true)
          .map((element) => element.absolute.path)
          .where((event) =>
              event.endsWith('.dart') &&
              (transformer.skipSource
                      .indexWhere((skip) => event.endsWith(skip)) ==
                  -1))
          .toList();
      includedPaths.addAll(files);
    }

    AnalysisContextCollection collection =
        new AnalysisContextCollection(includedPaths: includedPaths);
    await analyzeSomeFiles(collection, includedPaths);
    await Process.start('dartfmt', ['-w', 'lib/src/widgets']);
  }, timeout: Timeout(Duration(minutes: 5)));
  test('Get local flutter version', () {
    var jsonString = _getFlutterVersion();
    expect(jsonString['frameworkVersion'] != null, true,
        reason: "frameworkVersion is not found");
    expect(jsonString['channel'] != null, true, reason: "channel is not found");
    expect(jsonString['frameworkRevision'] != null, true,
        reason: "frameworkRevision is not found");
    expect(jsonString['engineRevision'] != null, true,
        reason: "engineRevision is not found");
    expect(jsonString['dartSdkVersion'] != null, true,
        reason: "dartSdkVersion is not found");
    print(jsonString);
  });
}

_getFlutterVersion() {
  var output = Process.runSync('flutter', ['--version', '--machine']);
  var jsonString = jsonDecode(output.stdout.toString());
  return jsonString;
}

analyzeSomeFiles(
    AnalysisContextCollection collection, List<String> includedPaths) async {
  const VERSION = '0.0.1';
  var flutter = _getFlutterVersion();
  var flutterVersion = flutter['frameworkVersion'];
  var dartVersion = flutter['dartSdkVersion'];

  StringBuffer buffer = StringBuffer('''
  // This file is generated by Fair, do not edit manually!
  // Updated on ${DateTime.now()}
  
  const fairVersion = '$VERSION';
  const flutterVersion = '$flutterVersion';
  const dartVersion = '$dartVersion';
  ''');

  var widgets = <Component>[];
  StringBuffer bindingBuffer = StringBuffer('''
  // This file is generated by Fair, do not edit manually!
  // Updated on ${DateTime.now()}
  
  import '../binding.dart';
  ''');
  var imports = <String>[];
  var components = <ComponentParts>[];
  for (String path in includedPaths) {
    AnalysisContext context = collection.contextFor(path);
    var w = await analyzeSingleFile(context, path);
    if (w == null || w.components.isEmpty) continue;
    widgets.addAll(w.components);
    var n = generateFileName(path);
    // name shrinking
    var index = n.indexOf('_');
    if (index != -1) {
      var s = shrinking.entries.firstWhere(
          (element) => n.startsWith(element.key),
          orElse: () => MapEntry('a', 'a'));
      n = s.value;
    }
    w.aliasGroup = n;
    components.add(w);
  }

  components.sort((a, b) => a.aliasGroup.compareTo(b.aliasGroup));
  shrinking.forEach((key, value) {
    var c = components.where((element) => element.aliasGroup == value).toList();
    if (c.isEmpty) return;
    var file = File('lib/src/widgets/\$\$$value.dart');
    imports.add('\$\$$value');
    var isCupertino = c.first.isCupertino;
    StringBuffer buffer = StringBuffer('''
  // This file is generated by Fair, do not edit manually!

  import 'package:flutter/${isCupertino ? "cupertino.dart" : "material.dart"}';
  ''');

    Set<String> allImports = c.fold(<String>{}, (p, e) {
      if (e.imports != null) {
        p.addAll(e.imports);
      }
      return p;
    });
    allImports.forEach((element) => buffer.write('import \'$element\';\n'));

    Set<String> l = c.fold(<String>{}, (p, e) {
      if (e.lines != null) {
        p.addAll(e.lines);
      }
      return p;
    });
    var lines = l.toList();
    lines.sort((a, b) =>
        a.startsWith('import') ? -1 : (b.startsWith('import') ? 1 : 0));
    lines.forEach((element) => buffer.write('$element\n'));
    Set<String> body = c.fold(<String>{}, (p, e) {
      if (e.body != null) {
        p.add(e.body);
      }
      return p;
    });
    buffer.write('''
  var p = () => {
  ''');
    body.forEach((element) => buffer.write('$element\n'));
    buffer.write('};');
    file.writeAsStringSync(buffer.toString());
  });

  for (var i = 0; i < imports.length; i++) {
    var n = imports[i];
    n = n.replaceAll('\$', '\\\$');
    bindingBuffer.write('import \'$n.dart\' as \$$i;');
  }

  bindingBuffer.write('''
  mixin \$BindingImpl implements Binding {
    final provider = [
  ''');

  for (var i = 0; i < imports.length; i++) {
    bindingBuffer.write('\$$i.p,');
  }

  bindingBuffer.write('];}');
  await File('lib/src/widgets/all.dart')
      .writeAsString(bindingBuffer.toString());
  var widgetCount = widgets.where((element) => element.isWidget).length;
  var apiCount = widgets.length - widgetCount;
  buffer.write('''
  const widgetCount = $widgetCount;
  const apiCount = $apiCount;
  const widgetNames = {
  ''');
  widgets.forEach((element) {
    buffer.write('\'${element.name}\': ${element.isWidget},');
  });
  buffer.write('};');
  await File('lib/src/widgets/version.dart').writeAsString(buffer.toString());
}

Future<ComponentParts> analyzeSingleFile(
    AnalysisContext context, String path) async {
  AnalysisSession session = context.currentSession;
  return await processFile(session, path);
}

Future<ComponentParts> processFile(AnalysisSession session, String path) async {
  UnitElementResult result = await session.getUnitElement(path);
  CompilationUnitElement element = result.element;
  var exposedAPI = printMembers(element);
  var count = exposedAPI?.length ?? 0;
  print('😀 $count widgets found inside $path');
  var isCupertino = path.contains('/cupertino/');

  StringBuffer buffer = StringBuffer();
  var clzName = basename(path).replaceAll('.dart', '');
  var imports = transformer.getImports(clzName);
  var lines = transformer.getLines(clzName, isCupertino);
  // constructor api
  var hasConstructor = false;
  exposedAPI?.forEach((e) {
    var hasNamedConstructor = false;
    Map<String, dynamic> defaultCache = {};
    e.constructor.forEach((element) {
      if (element.parameters != null && element.parameters.isNotEmpty) {
        hasNamedConstructor = true;
      }
      defaultCache =
          writeConstructor(buffer, element.name, element, defaultCache);
    });
    if (hasNamedConstructor) {
      e.fields.forEach((element) {
        ConstField field = element;
        buffer.write(
            '\'${field.name}.${field.field}\': ${field.name}.${field.field},');
      });
    } else if (e.fields.isNotEmpty) {
      var name = e.name;
      buffer.write('\'$name\': {');
      e.fields.forEach((e) {
        ConstField field = e;
        buffer.write('\'${field.field}\': ${field.name}.${field.field},');
      });
      buffer.write('},');
    }
  });
  if (!hasConstructor && exposedAPI != null && exposedAPI.isEmpty) return null;
  return ComponentParts(
    exposedAPI
        ?.map((e) => Component(
              e.name,
              e.constructor != null && e.constructor.length > 0
                  ? e.constructor[0].isWidget
                  : e.fields[0].isWidget,
            ))
        ?.toList(growable: false),
    buffer.toString(),
    imports: imports,
    lines: lines,
    isCupertino: isCupertino,
    clzName: clzName,
  );
}

class ComponentParts {
  final bool isCupertino;
  final List<String> imports;
  final List<String> lines;
  final List<Component> components;
  String aliasGroup;
  final String body;
  final String clzName;

  ComponentParts(
    this.components,
    this.body, {
    this.imports,
    this.lines,
    this.isCupertino,
    this.clzName,
  });
}

class Component {
  final String name;
  final bool isWidget;

  Component(this.name, this.isWidget);
}

String generateFileName(String path) {
  var fileName =
      path.split('packages/flutter/lib/src/')[1].replaceAll('/', '_');
  return fileName;
}

Map<String, dynamic> writeConstructor(StringBuffer buffer, String name,
    Constructor element, Map<String, dynamic> defaultCache) {
  buffer.write('\'$name\': (props) => $name(');
  if (element.parameters != null && element.parameters.isNotEmpty) {
    for (int i = 0; i < element.parameters.length; i++) {
      var p = element.parameters[i];
      if (!transformer.isPropSupported(p.name)) continue;
      String prop;
      // JSON属性中可能存在，浮点写成整型，需要兼容
      bool isDouble = p.type == 'double';
      if (p.defaultValueCode != null) {
        prop = isDouble
            ? 'props[\'${p.name}\']?.toDouble() ?? ${p.defaultValueCode},'
            : 'props[\'${p.name}\'] ?? ${p.defaultValueCode},';
        defaultCache[p.name] = p.defaultValueCode;
      } else if (defaultCache[p.name] != null) {
        prop = isDouble
            ? 'props[\'${p.name}\']?.toDouble() ?? ${defaultCache[p.name]},'
            : 'props[\'${p.name}\'] ?? ${defaultCache[p.name]},';
        print(
            '💕 using cached default value  ${element.name} ${p.name}=> ${defaultCache[p.name]}');
      } else {
        prop = isDouble
            ? 'props[\'${p.name}\']?.toDouble(),'
            : 'props[\'${p.name}\'],';
      }
      String namedDeclare = '${p.name}: $prop';
      var positionDeclare =
          isDouble ? 'props[\'pa\'][$i]?.toDouble(),' : 'props[\'pa\'][$i],';
      buffer.write(p.isNamed ? namedDeclare : positionDeclare);
    }
    var params = element.parameters.fold(
        '',
        (value, p) =>
            value + (p.isNamed ? '${p.type} ${p.name}, ' : '${p.name}, '));
    print('➡️ $name({$params})');
  } else {
    print('➡️ $name()');
  }
  buffer.write('),');
  return defaultCache;
}

class DemoAstVisitor extends GeneralizingAstVisitor<Map> {
  @override
  Map visitNode(AstNode node) {
    //输出遍历AST Node 节点内容
    print("${node.runtimeType}<---->${node.toSource()}");
    return super.visitNode(node);
  }
}

var transformer = TransformProxy();

bool matchType(InterfaceType type, List<String> widgets) {
  var hit = widgets.indexWhere((element) => element == type.name) != -1;
  if (hit) return true;
  return type.superclass != null ? matchType(type.superclass, widgets) : false;
}

class ClassExposed extends Exposed {
  final List<Constructor> constructor;
  final List<ConstField> fields;

  ClassExposed(
    String name, {
    this.constructor,
    this.fields,
  }) : super(name);

  int get size => constructor.length + fields.length;
}

abstract class Exposed {
  final String name;

  Exposed(this.name);
}

class Constructor extends Exposed {
  final List<Parameter> parameters;
  final bool isWidget;

  Constructor(String name, {this.parameters, this.isWidget = true})
      : super(name);

  @override
  String toString() {
    return name;
  }
}

class ConstField extends Exposed {
  final String field;
  final bool isWidget;

  ConstField(String name, {@required this.field, this.isWidget = true})
      : super(name);

  @override
  String toString() {
    return name + "." + field;
  }
}

class Parameter {
  final String name;
  final String type;
  final bool isNamed;
  final bool isOptional;
  final String defaultValueCode;

  Parameter({
    @required this.name,
    @required this.type,
    this.isNamed = false,
    this.isOptional = true,
    this.defaultValueCode,
  });
}

bool inValidElement(Element element) {
  if (!element.isPublic ||
      element.hasDeprecated ||
      element.hasVisibleForTesting ||
      element.isSynthetic) {
    return true;
  }
  // hasVisibleForTesting 与 isVisibleForTesting都不能正确识别注解visibleForTesting
  var meta = element.metadata;
  var isVisibleForTesting = meta.isNotEmpty
      ? meta.where((element) =>
          element.isVisibleForTesting ||
          element.toSource() == '@visibleForTesting')
      : null;
  return isVisibleForTesting != null && isVisibleForTesting.isNotEmpty;
}

List<ClassExposed> printMembers(CompilationUnitElement unitElement) {
  List<ClassExposed> exposed = [];
  // 枚举与class不同
  var apis = [...unitElement.types, ...unitElement.enums];
  for (ClassElement classElement in apis) {
    if (classElement.isAbstract || inValidElement(classElement)) {
      print('skip ' + classElement.name);
      continue;
    }
    List<Constructor> constructors = [];
    var isWidget = classElement.type is InterfaceType
        ? matchType(classElement.type, baseWidget)
        : false;
    var isAPI = classElement.type is InterfaceType
        ? matchType(classElement.type, baseAPI)
        : false;
    if (isWidget || isAPI) {
      print('${classElement.name} 😀');
      print('constructors ➡️');
      for (ConstructorElement constructorElement in classElement.constructors) {
        if (!inValidElement(constructorElement)) {
          String name;
          if (constructorElement.name == null ||
              constructorElement.name == '') {
            name = classElement.name;
          } else {
            name = '${classElement.name}.${constructorElement.name}';
          }
          print(' $name');
          // params
          List<Parameter> parameters;
          if (constructorElement.parameters.isNotEmpty) {
            parameters = constructorElement.parameters
                .map((e) => Parameter(
                      type: e.type.name,
                      name: e.name,
                      isNamed: e.isNamed,
                      isOptional: e.isOptional,
                      defaultValueCode: e.defaultValueCode,
                    ))
                .toList(growable: false);
          }

          constructors.add(
              Constructor(name, parameters: parameters, isWidget: isWidget));
        }
      }
      print('constructors ⬅️️');
    }
    List<ConstField> fields = [];
    for (FieldElement fieldElement in classElement.fields) {
      if (!inValidElement(fieldElement) && fieldElement.isConst) {
        print('  ❤️${classElement.name}.${fieldElement.name}');
        // 复用class类型（可能不一致）
        fields.add(ConstField(classElement.name,
            field: fieldElement.name, isWidget: isWidget));
      }
    }
/*    for (PropertyAccessorElement accessorElement in classElement.accessors) {
      if (!accessorElement.isSynthetic) {
        print('  ${accessorElement.name}');
      }
    }
    for (MethodElement methodElement in classElement.methods) {
      if (!methodElement.isSynthetic) {
        print('  ${methodElement.name}');
      }
    }*/
    if (constructors.isNotEmpty || fields.isNotEmpty /*|| enums.isNotEmpty*/) {
      exposed.add(ClassExposed(
        classElement.name,
        constructor: constructors,
        fields: fields,
      ));
    }
  }
  return exposed;
}
