/*
 * Copyright (C) 2005-present, 58.com.  All rights reserved.
 * Use of this source code is governed by a BSD type license that can be
 * found in the LICENSE file.
 */

import 'dart:convert';
import 'dart:io';

import 'package:analyzer/analyzer.dart';
import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:flutter_test/flutter_test.dart';

import 'private_smaple.dart';

void main() {
  var encoder = JsonEncoder.withIndent('  ');
  void visitDeclaration(CompilationUnitMember unitMember) {
    if (unitMember is ClassDeclaration) {
      for (ClassMember classMember in unitMember.members) {
        if (classMember is MethodDeclaration) {
          print('  ${classMember.name}');
        } else if (classMember is FieldDeclaration) {
          for (VariableDeclaration field in classMember.fields.variables) {
            print('  ${field.name.name}');
          }
        } else if (classMember is ConstructorDeclaration) {
          if (classMember.name == null) {
            print('  ${unitMember.name.name}');
          } else {
            print('  ${unitMember.name.name}.${classMember.name.name}');
          }
        }
      }
    }
  }

  String build2DSL(ClassDeclaration unitMember) {
    print(unitMember.name.name);
    var m = unitMember.members.where((classMember) =>
        classMember is MethodDeclaration &&
        classMember.name.name == 'build' &&
        classMember.returnType.type.name == 'Widget');
    if (m.isEmpty) {
      return null;
    }
    var buildFunc = m.first as MethodDeclaration;
    var statements = (buildFunc.body as BlockFunctionBody).block.statements;
    // TODO 直接return的情况
    if (statements.length == 1) {
      var instance = ((statements[0] as ReturnStatement).expression
          as InstanceCreationExpression);
      var data = encoder.convert(instance.accept(BuildVisitor()));
      print('===================\n$data');
      return data;
    }
    return null;
  }

  test('Test generate example/assets/bundle/listview.json', () async {
    var expect = File('example/assets/bundle/listview.json').readAsStringSync();
    var result = encoder.convert(jsonDecode(expect));
    assert(expect == result, "json string are not matched");

    var includedPaths = <String>[
      '/Users/aven/Desktop/fair/example/lib/page/sample_listview_page.dart'
    ];

    AnalysisContextCollection collection =
        new AnalysisContextCollection(includedPaths: includedPaths);
    includedPaths.forEach((e) async {
      await collection
          .contextFor(e)
          .currentSession
          .getResolvedUnit(e)
          .then((value) => value.unit.declarations)
          .then((unitMember) {
        unitMember
            .where((element) => element is ClassDeclaration)
            .map((e) => build2DSL(e))
            .forEach((element) {
          assert(expect == element || element == null,
              "json string are not matched");
        });
      });
    });
  });
  test('Test generate example/assets/bundle/video_card.json', () async {
    var expect =
        File('example/assets/bundle/video_card.json').readAsStringSync();
    var result = encoder.convert(jsonDecode(expect));
    assert(expect == result, "json string are not matched");

    var includedPaths = <String>[
      '/Users/aven/Desktop/fair/example/lib/page/zhihu/video_card.dart'
    ];

    AnalysisContextCollection collection =
        new AnalysisContextCollection(includedPaths: includedPaths);
    includedPaths.forEach((e) async {
      await collection
          .contextFor(e)
          .currentSession
          .getResolvedUnit(e)
          .then((value) => value.unit.declarations)
          .then((unitMember) {
        unitMember
            .where((element) => element is ClassDeclaration)
            .map((e) => build2DSL(e))
            .forEach((element) {
          assert(expect == element || element == null,
              "json string are not matched");
        });
      });
    });
  });
}

class PrintVisitor extends GeneralizingAstVisitor {
  @override
  Map visitNode(AstNode node) {
    print(" ${node.runtimeType} \t=>${node.toSource()}\n");
    return super.visitNode(node);
  }
}

class BuildVisitor extends MyAstVisitor {
  int i = 0;

  @override
  Map visitMethodDeclaration(MethodDeclaration node) {
    print("$i ${node.runtimeType} \t=>${node.toSource()}\n");
    return _buildMethodDeclaration(_safelyVisitNode(node.name),
        _safelyVisitNode(node.body), _safelyVisitNode(node.returnType),
        isAsync: node.body.isAsynchronous);
  }

  var directiveValue = {
    'NumericLiteral',
    'DoubleLiteral',
    'StringLiteral',
    'ListLiteral'
  };

  dynamic extractData(Map value) {
    if (value == null) return null;
    var type = value['type'];
    if ('Identifier' == type) {
      return '#(${value['name']})';
    } else if ('PrefixedIdentifier' == type) {
      var className = value['prefix']['name'];
      var constField = value['identifier']['name'];
      return "#($className.$constField)";
    } else if (directiveValue.contains(type)) {
      return value['value'];
    } else {
      return value;
    }
  }

  @override
  visitNamedExpression(NamedExpression node) {
    var key = _safelyVisitNode(node.name)['name'];
    var value = _safelyVisitNode(node.expression);
    return {key: extractData(value)};
  }

  @override
  Map visitSimpleIdentifier(SimpleIdentifier node) {
    return {"type": "Identifier", "name": node.name};
  }

  @override
  Map visitDoubleLiteral(DoubleLiteral node) {
    return {"type": 'DoubleLiteral', "value": node.value};
  }

  @override
  Map visitInstanceCreationExpression(InstanceCreationExpression node) {
    print("$i ${node.runtimeType} \t=>${node.toSource()}\n");
    var na = node.argumentList.arguments
        .where((element) => element is NamedExpression);
    var pa = node.argumentList.arguments
        .where((element) => !(element is NamedExpression));

    var data = <String, dynamic>{"className": node.constructorName.toString()};
    if (na.isNotEmpty) {
      data['na'] = na
          .map((e) => _safelyVisitNode(e))
          .reduce((value, element) => value..addAll(element));
    }
    if (pa.isNotEmpty) {
      data['pa'] = pa.map((e) => extractData(_safelyVisitNode(e))).toList();
    }
    return data;
  }

  Map _buildVisitListLiteral(List<Map> literal) =>
      {"type": "ListLiteral", "value": literal};

  @override
  Map visitListLiteral(ListLiteral node) {
    return _buildVisitListLiteral(_safelyVisitNodeList(node.elements));
  }

  Map _safelyVisitNode(AstNode node) {
    if (node != null) {
      return node.accept(this);
    }
    return null;
  }

  Map _buildMethodDeclaration(Map id, Map body, Map returnType,
          {bool isAsync: false}) =>
      {
        "type": "MethodDeclaration",
        "id": id,
        "body": body,
        "returnType": returnType,
      };

  /// 遍历节点列表
  List<Map> _safelyVisitNodeList(NodeList<AstNode> nodes) {
    List<Map> maps = [];
    if (nodes != null) {
      int size = nodes.length;
      for (int i = 0; i < size; i++) {
        var node = nodes[i];
        if (node != null) {
          var res = node.accept(this);
          if (res != null) {
            maps.add(res);
          }
        }
      }
    }
    return maps;
  }
}
